import socketserver
import subprocess
import time
import logging
import argparse
from threading import Thread
import socket
import sys
import json
import os
import re
from os import path

from sysloghandler import SyslogHandler

HOST = '0.0.0.0'
MAX_LENGTH = 4096
PORT = 10000
POLL_INTERVAL = 0.5
LOG_FILE_PATH = '/var/log/cyops/cyops-integrations/syslog/listener.log'

APPLIANCE_PRIVATE_KEY = ''
APPLIANCE_PUBLIC_KEY = ''
BASE_PATH = '/opt/cyops-integrations/integrations'
APPLIANCE_PATH = path.join(BASE_PATH, 'configs')

logging.basicConfig(filename=LOG_FILE_PATH, level=logging.DEBUG,
                    format='%(asctime)s %(levelname)s %(module)s %(funcName)s(): %(message)s')

logger = logging.getLogger(__name__)

servers = {}


def get_key(transport, port):
    return transport.upper() + '-' + str(port)


def get_server(transport, port):
    if 'UDP' == transport.upper():
        server = socketserver.UDPServer((HOST, port), SyslogHandler)
    elif 'TCP' == transport.upper():
        server = socketserver.TCPServer((HOST, port), SyslogHandler)
    else:
        raise Exception('Unsupported protocol')
    servers[get_key(transport, port)] = server
    return server


def start_server(transport, port, trigger, filter_str):
    # for idempotence
    if get_key(transport, port) in servers.keys():
        message = 'Server already up for transport: %s, port: %s, trigger: %s'% (transport, port, trigger)
        logger.info(message)
        return 0, message

    server = get_server(transport, port)
    server.trigger = trigger
    server.transport = transport
    server.filter_str = filter_str.strip('"')
    server.appliance_public_key = APPLIANCE_PUBLIC_KEY
    server.appliance_private_key = APPLIANCE_PRIVATE_KEY
    logger.info('Starting server for transport: %s, port: %s, trigger: %s'
                % (transport, port, trigger))
    Thread(target=server.serve_forever, args=(POLL_INTERVAL,)).start()
    message = 'Server started for transport: %s, port: %s, trigger: %s'\
              % (transport, port, trigger)
    logger.info(message)
    return 0, message


def stop_server(transport, port):
    server_key = get_key(transport, port)
    message = ''
    if server_key in servers:
        server = servers[server_key]
        Thread(target=server.shutdown).start()
        servers.pop(server_key)
        message = 'Server stopped for %s:%s' % (transport, str(port))
    else:
        message = 'No listener running for %s:%s' % (transport, str(port))
    logger.info(message)
    return 0, message


def restart_server(transport, port, trigger, filter_str):
    stop_server(transport, port)
    # confirm
    count = 0
    while count <= 10:
        try:
            pid = subprocess.check_output(['/usr/sbin/lsof', '-ti' + transport.upper() + ':' + str(port)])
            logger.info('pid for transport %s and port %s: "%s"' % (transport, str(port), str(pid)))
            # wait for stop
            logging.info('Waiting for process to stop: %s' % str(pid))
            time.sleep(5)
            count += 1
        except subprocess.CalledProcessError as e:
            logger.info('Did not find a running process for %s:%s' % (transport, str(port)))
            status, message = start_server(transport, port, trigger, filter_str)
            if status == 0:
                return 0, 'Server restarted for transport: %s, port: %s, trigger: %s'\
              % (transport, port, trigger)
    message = 'Timed out waiting for process %s to stop' % (str(pid))
    logger.error(message)
    return -1, message


def check_health(transport, port):
    try:
        pid = subprocess.check_output(['/usr/sbin/lsof', '-ti' + transport.upper() + ':' + str(port)])
        message = 'found running process for %s:%s with pid: %s' % (transport, str(port), str(pid))
        logger.info(message)
        return 0, message
    except subprocess.CalledProcessError as e:
        message = 'Did not find a running process for %s:%s' % (transport, str(port))
        logger.error(message)
        return -1, message


def shutdown_server():
    logger.info('Stopping all servers')
    for server in servers.values():
        Thread(target=server.shutdown, args=()).start()
    serversocket.close()
    logger.info('Sleeping for 5 seconds')
    time.sleep(5)
    os._exit(0)


def handle(clientsocket):
    payload_bytes = clientsocket.recv(MAX_LENGTH)
    if payload_bytes:
        payload = payload_bytes.decode('utf-8')
        parser = argparse.ArgumentParser(description='Syslog Server Actions')
        parser.add_argument('--start', help='Start Server', action='store_true', default=False, required=False)
        parser.add_argument('--stop', help='Stop Server', action='store_true', default=False, required=False)
        parser.add_argument('--restart', help='Restart Server', action='store_true', default=False, required=False)
        parser.add_argument('--check', help='Check if listener is running', action='store_true', default=False, required=False)
        parser.add_argument('--exit', help='Stop the socket server', action='store_true', default=False, required=False)
        parser.add_argument('--transport', help='Protocol to be used (TCP/UDP)', required=False)
        parser.add_argument('--port', help='Port for server', required=False)
        parser.add_argument('--trigger', help='CyOps endpoint to forward syslog to', required=False)
        parser.add_argument('--filter', help='Pattern to match before forwarding to CyberSponse', required=False)

        status = -1
        args_parsed = False
        try:
            # so that the main program does not exit if parsing fails
            args = parser.parse_args([p for p in re.split("( |\\\".*?\\\"|'.*?')", payload) if p.strip()])
            args_parsed = True
        except SystemExit as se:
            message = se.code
            logger.exception(se)

        if args_parsed:
            if args.exit:
                shutdown_server()
            try:
                if not args.port or not args.transport:
                    raise Exception('port and transport are mandatory arguments')
                port = int(args.port)
                transport = args.transport
                if args.stop:
                    status, message = stop_server(transport, port)
                elif args.check:
                    status, message = check_health(transport, port)
                else:
                    trigger = args.trigger
                    if not trigger:
                        raise Exception('trigger is a mandatory argument')
                    if args.start:
                        status, message = start_server(transport, port, trigger, args.filter)
                    elif args.restart:
                        status, message = restart_server(transport, port, trigger, args.filter)
                    else:
                        raise Exception('Unsupported function')
            except Exception as e:
                logger.exception(e)
                message = str(e)
        clientsocket.sendall(json.dumps({'status': status, 'message': message}).encode('utf-8'))
        clientsocket.close()

try:
    pri_key_path = os.path.join(APPLIANCE_PATH, 'APPLIANCE_PRIVATE_KEY')
    pub_key_path = os.path.join(APPLIANCE_PATH, 'APPLIANCE_PUBLIC_KEY')
    with open(pri_key_path, 'r') as private_key:
        APPLIANCE_PRIVATE_KEY = private_key.read().strip()
    with open(pub_key_path, 'r') as public_key:
        APPLIANCE_PUBLIC_KEY = public_key.read().strip()
except FileNotFoundError as e:
    logger.error('Failed to load authentication files')
    sys.exit()

serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
logger.info('Socket created')

# Bind socket to local host and port
try:
    serversocket.bind((HOST, PORT))
except socket.error as msg:
    logger.error('Bind failed.')
    sys.exit()

logger.info('Socket bind complete')

# Start listening on socket
serversocket.listen(10)
logger.info('Socket now listening')

# now keep talking with the client
while 1:
    # wait to accept a connection - blocking call
    try:
        (clientsocket, address) = serversocket.accept()
        handle(clientsocket)
    except KeyboardInterrupt:
        shutdown_server()

shutdown_server()
